==> stream-chat-android-client-5.3.0-sources.jar: io/getstream/chat/android/client/api/models/QuerySort.kt <==
/*
 * Copyright (c) 2014-2022 Stream.io Inc. All rights reserved.
 *
 * Licensed under the Stream License;
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    https://github.com/GetStream/stream-chat-android/blob/main/LICENSE
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.getstream.chat.android.client.api.models

import io.getstream.chat.android.client.api.models.QuerySort.SortAttribute.FieldSortAttribute
import io.getstream.chat.android.client.extensions.camelCaseToSnakeCase
import io.getstream.chat.android.client.extensions.snakeToLowerCamelCase
import io.getstream.chat.android.client.models.CustomObject
import io.getstream.chat.android.core.internal.InternalStreamChatApi
import io.getstream.logging.StreamLog
import kotlin.jvm.internal.Reflection
import kotlin.reflect.KClass
import kotlin.reflect.KProperty1
import kotlin.reflect.full.memberProperties

/**
 * Sort specification for api queries. You can specify QuerySort by referencing kotlin class property or passing field
 * name as string instance.
 * QuerySort.asc(Channel::memberCount) and QuerySort.asc<Channel>("member_count") mean the same.
 */
@Suppress("TooManyFunctions")
public class QuerySort<T : Any> {
    private val logger = StreamLog.getLogger("QuerySort")

    private var sortSpecifications: List<SortSpecification<T>> = emptyList()

    @InternalStreamChatApi
    /** Composite comparator based on sort attributes. */
    public val comparator: Comparator<in T>
        get() = CompositeComparator(sortSpecifications.map { it.comparator })

    private val SortSpecification<T>.comparator: Comparator<T>
        get() {
            return when (this.sortAttribute) {
                is FieldSortAttribute<T> -> this.sortAttribute.field.comparator(this.sortDirection)
                    .also { comparator ->
                        logger.d { "Returning field sort with name: ${sortAttribute.field.name}" }
                    }

                is SortAttribute.FieldNameSortAttribute -> this.sortAttribute.name.comparator(this.sortDirection)
                    .also { comparator ->
                        logger.d { "Returning field name sort comparator with name: ${sortAttribute.name}" }
                    }
            }
        }

    @Suppress("UNCHECKED_CAST")
    private fun KProperty1<T, Comparable<*>?>.comparator(sortDirection: SortDirection): Comparator<T> =
        this.let { compareProperty ->
            Comparator { c0, c1 ->
                compare(
                    (compareProperty.getter.call(c0) as? Comparable<Any>),
                    (compareProperty.getter.call(c1) as? Comparable<Any>),
                    sortDirection
                )
            }
        }

    @Suppress("UNCHECKED_CAST")
    private fun String.comparator(sortDirection: SortDirection): Comparator<T> =
        Comparator { o1, o2 ->
            compare(
                (o1.getMemberPropertyOrExtra(this) as? Comparable<Any>),
                (o2.getMemberPropertyOrExtra(this) as? Comparable<Any>),
                sortDirection
            )
        }

    private fun compare(first: Comparable<Any>?, second: Comparable<Any>?, sortDirection: SortDirection): Int {
        return when {
            first == null && second == null -> EQUAL_ON_COMPARISON
            first == null && second != null -> LESS_ON_COMPARISON * sortDirection.value
            first != null && second == null -> MORE_ON_COMPARISON * sortDirection.value
            first != null && second != null -> first.compareTo(second) * sortDirection.value
            else -> error("Impossible case!")
        }
    }

    private fun Any.getMemberPropertyOrExtra(name: String): Any? =
        name.snakeToLowerCamelCase().let { fieldName ->
            this::class.memberProperties
                .firstOrNull { it.name == fieldName }
                ?.also {
                    logger.d { "Could find a member for property: $fieldName" }
                }
                ?.getter
                ?.call(this)
                ?: (this as? CustomObject)?.extraData?.get(name).also { extraDataObject ->
                    buildString {
                        append("Could not find a member for property named: $name")
                        if (extraDataObject != null) {
                            append(", but a fields was found in extraData.")
                        } else {
                            append("and nothing was found in the extra data")
                        }
                    }.let { string -> logger.d { string } }
                }
        }

    private fun add(sortSpecification: SortSpecification<T>): QuerySort<T> {
        sortSpecifications = sortSpecifications + sortSpecification
        return this
    }

    public fun asc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> {
        return add(SortSpecification(FieldSortAttribute(field, field.name.camelCaseToSnakeCase()), SortDirection.ASC))
    }

    public fun desc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> {
        return add(SortSpecification(FieldSortAttribute(field, field.name.camelCaseToSnakeCase()), SortDirection.DESC))
    }

    public fun asc(fieldName: String, javaClass: Class<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, javaClass), SortDirection.ASC))
    }

    public fun desc(fieldName: String, javaClass: Class<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, javaClass), SortDirection.DESC))
    }

    public fun asc(fieldName: String): QuerySort<T> {
        return add(SortSpecification(SortAttribute.FieldNameSortAttribute(fieldName), SortDirection.ASC))
    }

    public fun desc(fieldName: String): QuerySort<T> {
        return add(SortSpecification(SortAttribute.FieldNameSortAttribute(fieldName), SortDirection.DESC))
    }

    public fun asc(fieldName: String, kClass: KClass<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, kClass), SortDirection.ASC))
    }

    public fun desc(fieldName: String, kClass: KClass<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, kClass), SortDirection.DESC))
    }

    public fun toDto(): List<Map<String, Any>> = sortSpecifications.map { sortSpec ->
        listOf(KEY_FIELD_NAME to sortSpec.sortAttribute.name, KEY_DIRECTION to sortSpec.sortDirection.value).toMap()
    }

    public fun toList(): List<Pair<String, SortDirection>> =
        sortSpecifications.map { it.sortAttribute.name to it.sortDirection }

    private fun getSortFeature(fieldName: String, javaClass: Class<T>): SortAttribute<T> {
        @Suppress("UNCHECKED_CAST")
        val kClass = Reflection.createKotlinClass(javaClass) as KClass<T>
        return getSortFeature(fieldName, kClass)
    }

    private fun getSortFeature(fieldName: String, kClass: KClass<T>): SortAttribute<T> {
        return kClass.members.filterIsInstance<KProperty1<T, Comparable<*>?>>()
            .firstOrNull { it.name == fieldName.snakeToLowerCamelCase() }
            ?.let { FieldSortAttribute(it, fieldName) }
            .also { fieldSortAttribute ->
                logger.d { "[getSortFeature] A field to sort was found. Using field: $fieldSortAttribute" }
            }
            ?: SortAttribute.FieldNameSortAttribute<T>(fieldName)
                .also { fieldNameSortAttribute ->
                    logger.d {
                        "[getSortFeature] A field to sort was NOT found. Using field by name: $fieldNameSortAttribute"
                    }
                }
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as QuerySort<*>

        if (sortSpecifications != other.sortSpecifications) return false

        return true
    }

    override fun hashCode(): Int {
        return sortSpecifications.hashCode()
    }

    override fun toString(): String {
        return sortSpecifications.toString()
    }

    private data class SortSpecification<T>(
        val sortAttribute: SortAttribute<T>,
        val sortDirection: SortDirection,
    )

    /** Inner representation of sorting feature specification. */
    private sealed class SortAttribute<T> {
        /** Name of attribute */
        abstract val name: String

        /** KProperty referenced attribute. */
        data class FieldSortAttribute<T>(val field: KProperty1<T, Comparable<*>?>, override val name: String) :
            SortAttribute<T>()

        /** Referenced by name attribute. */
        data class FieldNameSortAttribute<T>(override val name: String) : SortAttribute<T>()
    }

    /** Sort order which can be ascending or descending. */
    public enum class SortDirection(public val value: Int) {
        /** Descending sort order. */
        DESC(-1),

        /** Ascending sort order. */
        ASC(1)
    }

    internal class CompositeComparator<T>(private val comparators: List<Comparator<T>>) : Comparator<T> {
        override fun compare(o1: T, o2: T): Int =
            comparators.fold(EQUAL_ON_COMPARISON) { currentComparisonValue, comparator ->
                when (currentComparisonValue) {
                    EQUAL_ON_COMPARISON -> comparator.compare(o1, o2)
                    else -> currentComparisonValue
                }
            }
    }

    public companion object {
        public const val KEY_DIRECTION: String = "direction"
        public const val KEY_FIELD_NAME: String = "field"
        private const val MORE_ON_COMPARISON = 1
        private const val EQUAL_ON_COMPARISON = 0
        private const val LESS_ON_COMPARISON = -1

        public inline fun <reified T : Any> QuerySort<T>.ascByName(fieldName: String): QuerySort<T> =
            asc(fieldName, T::class)

        public inline fun <reified T : Any> QuerySort<T>.descByName(fieldName: String): QuerySort<T> =
            desc(fieldName, T::class)

        public inline fun <reified T : Any> asc(fieldName: String): QuerySort<T> = QuerySort<T>().ascByName(fieldName)
        public inline fun <reified T : Any> desc(fieldName: String): QuerySort<T> = QuerySort<T>().descByName(fieldName)
        public fun <T : Any> asc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> = QuerySort<T>().asc(field)
        public fun <T : Any> desc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> = QuerySort<T>().desc(field)
    }
}

==> stream-chat-android-client-5.3.1-sources.jar: io/getstream/chat/android/client/api/models/QuerySort.kt <==
/*
 * Copyright (c) 2014-2022 Stream.io Inc. All rights reserved.
 *
 * Licensed under the Stream License;
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    https://github.com/GetStream/stream-chat-android/blob/main/LICENSE
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.getstream.chat.android.client.api.models

import io.getstream.chat.android.client.api.models.QuerySort.SortAttribute.FieldSortAttribute
import io.getstream.chat.android.client.api.models.internal.CompositeComparator
import io.getstream.chat.android.client.extensions.camelCaseToSnakeCase
import io.getstream.chat.android.client.models.CustomObject
import io.getstream.chat.android.core.internal.InternalStreamChatApi
import io.getstream.logging.StreamLog
import kotlin.jvm.internal.Reflection
import kotlin.reflect.KClass
import kotlin.reflect.KProperty1

/**
 * Sort specification for api queries. You can specify QuerySort by referencing kotlin class property or passing field
 * name as string instance.
 * QuerySort.asc(Channel::memberCount) and QuerySort.asc<Channel>("member_count") mean the same.
 */
@Suppress("TooManyFunctions")
public class QuerySort<T : Any> {
    private val logger = StreamLog.getLogger("QuerySort")

    private var sortSpecifications: List<SortSpecification<T>> = emptyList()
    private val fieldSearcher: FieldSearcher = FieldSearcher()

    @InternalStreamChatApi
    /** Composite comparator based on sort attributes. */
    public val comparator: Comparator<in T>
        get() = CompositeComparator(sortSpecifications.map { it.comparator })

    private val SortSpecification<T>.comparator: Comparator<T>
        get() {
            return when (this.sortAttribute) {
                is FieldSortAttribute<T> -> this.sortAttribute.field.comparator(this.sortDirection)

                is SortAttribute.FieldNameSortAttribute -> this.sortAttribute.name.comparator(this.sortDirection)
            }
        }

    private fun compare(first: Comparable<Any>?, second: Comparable<Any>?, sortDirection: SortDirection): Int {
        return when {
            first == null && second == null -> EQUAL_ON_COMPARISON
            first == null && second != null -> LESS_ON_COMPARISON * sortDirection.value
            first != null && second == null -> MORE_ON_COMPARISON * sortDirection.value
            first != null && second != null -> first.compareTo(second) * sortDirection.value
            else -> error("Impossible case!")
        }
    }

    private fun add(sortSpecification: SortSpecification<T>): QuerySort<T> {
        sortSpecifications = sortSpecifications + sortSpecification
        return this
    }

    public fun asc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> {
        return add(SortSpecification(FieldSortAttribute(field, field.name.camelCaseToSnakeCase()), SortDirection.ASC))
    }

    public fun desc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> {
        return add(SortSpecification(FieldSortAttribute(field, field.name.camelCaseToSnakeCase()), SortDirection.DESC))
    }

    public fun asc(fieldName: String, javaClass: Class<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, javaClass), SortDirection.ASC))
    }

    public fun desc(fieldName: String, javaClass: Class<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, javaClass), SortDirection.DESC))
    }

    public fun asc(fieldName: String): QuerySort<T> {
        return add(SortSpecification(SortAttribute.FieldNameSortAttribute(fieldName), SortDirection.ASC))
    }

    public fun desc(fieldName: String): QuerySort<T> {
        return add(SortSpecification(SortAttribute.FieldNameSortAttribute(fieldName), SortDirection.DESC))
    }

    public fun asc(fieldName: String, kClass: KClass<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, kClass), SortDirection.ASC))
    }

    public fun desc(fieldName: String, kClass: KClass<T>): QuerySort<T> {
        return add(SortSpecification(getSortFeature(fieldName, kClass), SortDirection.DESC))
    }

    public fun toDto(): List<Map<String, Any>> = sortSpecifications.map { sortSpec ->
        listOf(KEY_FIELD_NAME to sortSpec.sortAttribute.name, KEY_DIRECTION to sortSpec.sortDirection.value).toMap()
    }

    public fun toList(): List<Pair<String, SortDirection>> =
        sortSpecifications.map { it.sortAttribute.name to it.sortDirection }

    private fun getSortFeature(fieldName: String, javaClass: Class<T>): SortAttribute<T> {
        @Suppress("UNCHECKED_CAST")
        val kClass = Reflection.createKotlinClass(javaClass) as KClass<T>
        return getSortFeature(fieldName, kClass)
    }

    private fun getSortFeature(fieldName: String, kClass: KClass<T>): SortAttribute<T> {
        return fieldSearcher.findComparableMemberProperty(fieldName, kClass)
            ?.let { FieldSortAttribute(it, fieldName) }
            .also { fieldSortAttribute ->
                logger.d { "[getSortFeature] A field to sort was found. Using field: $fieldSortAttribute" }
            }
            ?: SortAttribute.FieldNameSortAttribute(fieldName)
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as QuerySort<*>

        if (sortSpecifications != other.sortSpecifications) return false

        return true
    }

    override fun hashCode(): Int {
        return sortSpecifications.hashCode()
    }

    override fun toString(): String {
        return sortSpecifications.toString()
    }

    private data class SortSpecification<T>(
        val sortAttribute: SortAttribute<T>,
        val sortDirection: SortDirection,
    )

    @Suppress("UNCHECKED_CAST")
    private fun KProperty1<T, Comparable<*>?>.comparator(sortDirection: SortDirection): Comparator<T> =
        this.let { compareProperty ->
            Comparator { c0, c1 ->
                compare(
                    (compareProperty.getter.call(c0) as? Comparable<Any>),
                    (compareProperty.getter.call(c1) as? Comparable<Any>),
                    sortDirection
                )
            }
        }

    @Suppress("UNCHECKED_CAST")
    private fun String.comparator(sortDirection: SortDirection): Comparator<T> =
        Comparator { o1, o2 ->
            compare(
                comparableFromExtraData(o1, this) ?: fieldSearcher.findComparable(o1, this),
                comparableFromExtraData(o2, this) ?: fieldSearcher.findComparable(o2, this),
                sortDirection
            )
        }

    private fun comparableFromExtraData(any: Any, field: String): Comparable<Any>? {
        return (any as? CustomObject)?.extraData?.get(field) as? Comparable<Any>
    }

    /** Inner representation of sorting feature specification. */
    private sealed class SortAttribute<T> {
        /** Name of attribute */
        abstract val name: String

        /** KProperty referenced attribute. */
        data class FieldSortAttribute<T>(val field: KProperty1<T, Comparable<*>?>, override val name: String) :
            SortAttribute<T>()

        /** Referenced by name attribute. */
        data class FieldNameSortAttribute<T>(override val name: String) : SortAttribute<T>()
    }

    /** Sort order which can be ascending or descending. */
    public enum class SortDirection(public val value: Int) {
        /** Descending sort order. */
        DESC(-1),

        /** Ascending sort order. */
        ASC(1)
    }

    public companion object {
        public const val KEY_DIRECTION: String = "direction"
        public const val KEY_FIELD_NAME: String = "field"
        private const val MORE_ON_COMPARISON = 1
        internal const val EQUAL_ON_COMPARISON = 0
        private const val LESS_ON_COMPARISON = -1

        public inline fun <reified T : Any> QuerySort<T>.ascByName(fieldName: String): QuerySort<T> =
            asc(fieldName, T::class)

        public inline fun <reified T : Any> QuerySort<T>.descByName(fieldName: String): QuerySort<T> =
            desc(fieldName, T::class)

        public inline fun <reified T : Any> asc(fieldName: String): QuerySort<T> = QuerySort<T>().ascByName(fieldName)
        public inline fun <reified T : Any> desc(fieldName: String): QuerySort<T> = QuerySort<T>().descByName(fieldName)
        public fun <T : Any> asc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> = QuerySort<T>().asc(field)
        public fun <T : Any> desc(field: KProperty1<T, Comparable<*>?>): QuerySort<T> = QuerySort<T>().desc(field)
    }
}
